!+
! Module: createpseudoevt_mod
!
! Authors:  M.T. Ceballos -  SSC (IFCA) 
!           F.J. Carrera -  SSC (IFCA)  
!           X. Barcons - SSC (IFCA)
!          
! Description: Creates the pseudoEvt file with X/Y and THETA columns
!              from input image
!              
!

module createpseudoevt_mod

  use dal
  use cal
  use my_psf_mod
  use coordinate_mod
  use param  
  use errorHandling
  
  implicit none 
  
  private 

  public createpseudoevt

contains 

  !---------------------------------------------------------------------
  subroutine createpseudoevt
  !---------------------------------------------------------------------
    implicit none


    real(kind=single) :: tcrpxX,tcrpxY,pi=3.1415926536E0
    real(kind=single) :: crpix1,crpix2,crval1,crval2,cdelt1,cdelt2
    real(kind=single) :: tcdltX,tcdltY
    real(kind=single),dimension(:),allocatable :: xPixelij,yPixelij
    real(kind=single), dimension(:,:), pointer :: pr32_inImage

    real(kind=double),dimension(:),allocatable :: RaDeg,DecDeg
    real(kind=double), dimension(:), pointer :: coX_Data,coY_Data
    real(kind=double), dimension(:,:), pointer :: p_inImage
    real(kind=double), dimension(:), pointer :: cotheta_Data
    real(kind=double),dimension(:),allocatable :: theta
    real(kind=double) :: sincrota2,coscrota2,crota2
    real(kind=double) :: expStartTimeStamp,distRarefPixRad,distDecrefPixRad
    real(kind=double) :: distXrefPixRad,distYrefPixRad,raRad,decRad &
         ,refRaRad,refDecRad

    integer(kind=Int8),dimension(:,:), pointer :: pi8_inImage
    integer(kind=Int16) :: status
    integer(kind=Int16),dimension(:,:), pointer :: pi16_inImage
    integer(kind=Int32) :: i,j,ipix,type_inImage,nrows,instrumentID
    integer(kind=Int32),dimension(:), pointer :: shape_inImage
    integer(kind=Int32),dimension(:,:), pointer :: pi32_inImage


    character(len=256) :: outEvt,inImage,dateTimeStr
    
 
    logical :: alloc_ptr

    type(DataSetT) :: set_outEvt,set_inImage
    type(TableT):: tab_outEvt
    type(ArrayT) :: arr_inImage
    type(ColumnT) :: coX_outEvt,coY_outEvt,coTheta_outEvt
!    type(ColumnT) :: coDetX_outEvt,coDetY_outEvt
    type(SpacecraftAttitudeType) :: scAttitude

!   read input parameters from param file
    
    outEvt=stringParameter("set")
    inImage=stringParameter("imageset")
    tcdltX=realParameter("tcdltx")
    tcdltY=realParameter("tcdlty")
    tcrpxX=realParameter("tcrpxx")
    tcrpxY=realParameter("tcrpxy")
    
    set_inImage=dataSet(inImage,Read)
    arr_inImage=array(set_inImage,"PRIMARY",Read)
    type_inImage=ArrayDataType(arr_inImage)
    shape_inImage => dimensions(arr_inImage)

!
!   Reading some important IMAGE Attributes
!

    if(hasAttribute(set_inImage,"CDELT1"))then
       cdelt1=real32Attribute(set_inImage,"CDELT1")
       cdelt2=real32Attribute(set_inImage,"CDELT2")
       crpix1=real32Attribute(set_inImage,"CRPIX1")
       crpix2=real32Attribute(set_inImage,"CRPIX2")
       crval1=real32Attribute(set_inImage,"CRVAL1")
       crval2=real32Attribute(set_inImage,"CRVAL2")
    elseif(hasAttribute(arr_inImage,"CDELT1"))then
       cdelt1=real32Attribute(arr_inImage,"CDELT1")
       cdelt2=real32Attribute(arr_inImage,"CDELT2")
       crpix1=real32Attribute(arr_inImage,"CRPIX1")
       crpix2=real32Attribute(arr_inImage,"CRPIX2")
       crval1=real32Attribute(arr_inImage,"CRVAL1")
       crval2=real32Attribute(arr_inImage,"CRVAL2")
    else
      write(errstr,*)"No keyword CDELT1 in:",inImage
      call fatal('NonStandardInput')
    endif
    if(hasAttribute(set_inImage,"CROTA2"))then
       crota2=real64Attribute(set_inImage,"CROTA2")
    elseif(hasAttribute(arr_inImage,"CROTA2"))then
       crota2=real64Attribute(arr_inImage,"CROTA2")
    else
       crota2=0.d0
    endif

    if(stringAttribute(set_inImage,"INSTRUME") == "EMOS1")then
       instrumentID=4
    elseif(stringAttribute(set_inImage,"INSTRUME") == "EMOS2")then
       instrumentID=5
    elseif(stringAttribute(set_inImage,"INSTRUME") == "EPN")then
       instrumentID=6
    endif

    dateTimeStr=stringAttribute(set_inImage,"DATE-OBS")
    scAttitude%ra=real32Attribute(set_inImage,"RA_PNT")
    scAttitude%dec=real32Attribute(set_inImage,"DEC_PNT")
    scAttitude%apos=real32Attribute(set_inImage,"PA_PNT")

    expStartTimeStamp=toTimeTag(dateTimeStr)
    scAttitude%ra=scAttitude%ra*pi/180.
    scAttitude%dec=scAttitude%dec*pi/180.
    scAttitude%apos=scAttitude%apos*pi/180.

!   select pointer type for input Image

    alloc_ptr=.TRUE.
    select case (type_inImage)
    case(INTEGER8)
       pi8_inImage => int8Array2Data(arr_inImage)
       allocate(p_inImage(0:shape_inImage(0)-1,0:shape_inImage(1)-1))
       p_inImage = real(pi8_inImage, DOUBLE)
    case(INTEGER16)
       pi16_inImage => int16Array2Data(arr_inImage)
       allocate(p_inImage(0:shape_inImage(0)-1,0:shape_inImage(1)-1))
       p_inImage = real(pi16_inImage, DOUBLE)
    case(INTEGER32)
       pi32_inImage => int32Array2Data(arr_inImage)
       allocate(p_inImage(0:shape_inImage(0)-1,0:shape_inImage(1)-1))
       p_inImage = real(pi32_inImage, DOUBLE)
    case(REAL32)
       pr32_inImage => real32Array2Data(arr_inImage)
       allocate(p_inImage(0:shape_inImage(0)-1,0:shape_inImage(1)-1))
       p_inImage = real(pr32_inImage, DOUBLE)
    case(REAL64)
       p_inImage => real64Array2Data(arr_inImage)
       alloc_ptr=.FALSE.
    case DEFAULT
       call fatal('InvalidDataType','Invalid data &
            &type in input Image')
    end select

    call release(set_inImage)

    coscrota2=cos(pi/180.*crota2)
    sincrota2=sin(pi/180.*crota2)
    refRaRad=pi/180.d0*crval1
    refDecRad=pi/180.d0*crval2

    nrows=shape_inImage(0)*shape_inImage(1)
    allocate(xPixelij(0:nrows-1),yPixelij(0:nrows-1))
    allocate(RaDeg(0:nrows-1))
    allocate(DecDeg(0:nrows-1))
    allocate(theta(0:nrows-1))


    ipix=0
    do i=0,shape_inImage(0)-1
       do j=0,shape_inImage(1)-1
          xPixelij(ipix)=(i+1-crpix1)*cdelt1/tcdltX+ tcrpxX
          yPixelij(ipix)=(j+1-crpix2)*cdelt2/tcdltY+ tcrpxY

          distXrefPixRad=(i+1.5-crpix1)*cdelt1*pi/180. 
          distYrefPixRad=(j+1.5-crpix2)*cdelt2*pi/180. 
          distRarefPixRad=distXrefPixRad*coscrota2 - distYrefPixRad*sincrota2
          distDecrefPixRad=distXrefPixRad*sincrota2 + distYrefPixRad*coscrota2

          call antisp(distRarefPixRad,distDecrefPixRad,refRaRad,&
               refDecRad,raRad,decRad)
          raDeg(ipix)=raRad*180./pi
          decDeg(ipix)=decRad*180./pi
          ipix=ipix+1
       end do
    end do


    if(alloc_ptr)deallocate(p_inImage)

    set_outEvt=dataSet(outEvt,CREATE)
    tab_outEvt=addtable(set_outEvt,"EVENTS",nrows)
    coX_outEvt=addcolumn(tab_outEvt,"X",Real64) 
    coY_outEvt=addcolumn(tab_outEvt,"Y",Real64)
    coTheta_outEvt=addcolumn(tab_outEvt,"THETA",Real64) 
    coX_Data=>Real64Data(coX_outEvt)
    coY_Data=>Real64Data(coY_outEvt)
    cotheta_Data=>real64Data(coTheta_outEvt)
                                  !arcs: angular deviation from optical axis
    do i=0,nrows-1
       coX_Data(i)=xPixelij(i)
       coY_Data(i)=yPixelij(i)
    enddo

    deallocate(xPixelij,yPixelij)


    
    call CAL_setState(instrument=instrumentId)
!    call CAL_setState(inImage)
    
    status=0

    call raDecToDetXY(RaDeg, DecDeg, scAttitude, expStartTimeStamp&
      , myThetaArcsec=cotheta_Data, status=status)

    call release(set_outEvt);

  end subroutine createpseudoevt

  SUBROUTINE ANTISP(DXN,DYN,DNOMRA,DNOMDE,DRA1,DDE1)

    !C calculates RA, DEC from pseudo-cartesian coordinates XN, YN
    !C (Performs exactly the opposite operation as SPHERE)
    !
    !************************ FFORM VERSION 3.0 ********** 31-JAN-90 09:20
    !
    !A  author : GRH              
    !U  update : PTK              date: 06-DEC-1990 17:03
    !U  update : PTK              date: 08-JAN-1990 08:50 angles in radians
    !U  update : EXSAS              date: 03-AUG-1993 19:40
    !
    !T  status: not tested
    !
    !   general description:
    !G  calculates RA, DEC from pseudo-cartesian coordinates XN, YN
    !G  (Performs exactly the opposite operation as SPHERE)
    !
    !   call_var.          type I/O description
    !P  DXN                 D   i   pseudo cartesian x coordinate in radians
    !P  DYN                 D   i   pseudo cartesian y coordinate in radians
    !P  DNOMRA              D   i  reference right ascension (radians) 
    !P                             for DXN = 0 and DYN = 0.
    !P  DNOMDE              D   i  reference declination (radians)
    !P                             for DXN = 0 and DYN = 0.
    !P  DRA1                D   o actual right ascension for DXN, DYN (radians)
    !P  DDE1                D   o actual declination for DXN, DYN (radians)
    !P  IERR                I   O = 0 no error 
    !P                              = 3 
    !
    !   include_block_name          description
    !I  RCOMMON:CGENL.CMN           general parameter common block
    !
    !   routines_called    type     description
    !R
    !
    !   extensions/system calls     description
    !X  
    !
    !***********************************************************************
    !
    !   variables   meaning
    !
    IMPLICIT NONE
    !
    real(double) :: DINRA,DINDE,DXN1,DYN1
    real(double) :: DXN,DYN,DNOMRA,DNOMDE,DRA1,DDE1
    real(double),parameter :: PID=3.1415926535898
    
    DINRA=DNOMRA
    DINDE=DNOMDE
    DXN1=DXN 
    DYN1=DYN
    CALL DTRA(-DINDE,PID,0.D0,DXN1,DYN1,DRA1,DDE1)
    DRA1=DRA1+DNOMRA
    IF(DRA1.GT.2.*PID)DRA1=DRA1-2.*PID
  END SUBROUTINE ANTISP
  
  SUBROUTINE DTRA(DRD,D6,D8,D2,D1,D2N,D1N)                       
    !
    !C  coordinate transformation
    !
    !************************ FFORM VERSION 3.0 ********** 31-JAN-90 09:20
    !
    !A  author : ASTROLIB from GRH       
    !U  update : PTK            date: 06-DEC-1990 17:30
    !U  update : EXSAS              date: 06-AUG-1991 16:06
    !U  update : EXSAS              date: 03-AUG-1993 19:42
    !
    !T  status: not tested
    !
    !   general description:
    !G  coordinate transformation
    !
    !   call_var.          type I/O description
    !P  DRD                 R*8 
    !P  D6                  R*8
    !P  D8                  R*8
    !P  D2                  R*8
    !P  D1                  R*8
    !P  D2N                 R*8
    !P  IERR                I     O = 0 no error 
    !P                              = 3 
    !
    !   include_block_name          description
    !I  RCOMMON:CGENL.CMN           general parameter common block
    !
    !   routines_called    type     description
    !R
    !
    !   extensions/system calls     description
    !X  
    !
    !***********************************************************************
    !
    !   variables   meaning
    !
    IMPLICIT NONE
      !
    real(double) :: D2,D6,D1,DR,DRD,DSIN2,DCOS2,DSIN1,DCOS1
    real(double) :: DSINR,DCOSR,DX,D7,DY,D2N,D8,DDNS,DDNC
    real(double) :: D1N
    real(double),parameter :: DPI=3.1415926535898, DF=1.D0
    
    D2=D2-D6                          
    D1=D1*DF
    D2=D2*DF
    DR=DRD*DF
    DSIN2=DSIN(D2)
    DCOS2=DCOS(D2)
    DSIN1=DSIN(D1)
    DCOS1=DCOS(D1)
    DSINR=DSIN(DR)
    DCOSR=DCOS(DR)
    DX=-DSIN2*DCOS1
    D7=DCOS1*DCOS2
    DY=DSINR*DSIN1-D7*DCOSR
    D2N=DATAN2(DX,DY)
    D2N=D2N/DF
    
    IF(D2N < 0.D0)D2N=D2N+2.*DPI
    D2N=D8+D2N
    
    IF(D2N < 0.D0)D2N=D2N+2.*DPI
    
    IF(D2N > 2.*DPI)D2N=D2N-2.*DPI
    DDNS=D7*DSINR+DCOSR*DSIN1
    DDNC=DSQRT(1.D0-DDNS*DDNS)
    D1N=DATAN2(DDNS,DDNC)
    D1N=D1N/DF
    
  END SUBROUTINE DTRA
  










end module createpseudoevt_mod
