#!/usr/local/bin/perl
# 
# Usage :
#
# flspec sourcelistset="Name of the fits file containing the list of 
#                     observed sources" 
#        set="EVENT LIST filename of the EPIC imaging observation"
#        imageset="EPIC Raw Image in band " 
#        exposureset="EPIC Exposure Map corresponding to image"
#        tmpflatset="Name of the file containing the temporary 
#                     flatfielded image"
#        tmpsrcregionset="Name of the temporary file containing the regions 
#                     for the sources"
#        tmpbkgregionset="Name of the temporary file containing the regions 
#                     for the sources background"
#        tmpposmaskset="Name of the file containing the temporary positive 
#                      image mask"
#        tmpnegmaskset="Name of the file containing the temporary negative 
#                      image mask"
#        tmpposspecset="Name of the file containing a temporary positive 
#                      spectrum"
#        tmpnegspecset="Name of the file containing a temporary negative 
#                      spectrum"
#        tmpdiffset="Name of the file containing a temporary subtracted
#                      spectrum"
#        binsize="Size of the rebinned pixels in degrees to define 
#                 positive and negative fluctuations"
#        spectrumset="Output fits filename containing the spectrum of the 
#               background fluctuations at different offaxis angles" 
#     
#     
#
#   Read input command line options: 
#        sourcelistset=$srclist  set=$evtlist  imageset=$img 
#        exposureset=$expm tmpflatset=$flatimg tmpsrcregionset=$temp1 
#        tmpbkgregionset=$temp2
#        tmpposmaskset=$posit_mask tmpnegmaskset=negat_mask
#        tmpposspecset=$spec_pos tmpnegspecset=$spec_neg
#        tmpdiffset=$temp3 $binsize=$binsize 
#        spectrumset=$flspecfile
#        
#        
#
#
# ------------------------------------------------------
# Define variables   
# ------------------------------------------------------ 

use DAL;
use SAS;
use strict;

sub flspec{
  
  my $binsize;                # deg for binned image pixel size
  my $rmax=12.0;              # arcm : maximum radius for annuli
  my $nann=10;                # number of annuli
  my $XcolMask="X";           # columns to extract spectra
  my $YcolMask="Y";           # 
  my ($ar,$at,$att,$co,$crpix1,$crpix2,$coXIn,$coYIn,$coXPseudo,$coYPseudo);
  my ($ds,$evtlist,$evt,$expm,$expr,$exposds,$expostb,$evIn,$evPseudo);
  my ($factor,$factorFormat,$filter,$filterexp,$flatimg,$flspecfile);
  my ($hascol,$ids,$img,$inPixSize,$jj,$jj1,$key,$label);
  my ($nspec,$negat1,$negat_mask,$msg); 
  my ($off_arcs,$pattern,$posit_mask,$pp,$posit1,$radius_int,$rowidStr);
  my ($specBin,$SolidAngle,$srclist,$temp3,$temp31,$spec_pos,$spec_neg);
  my ($specChanMax,$SolidAngle_pos,$SolidAngle_neg,$SolidAngleFormat,$spec_1);
  my ($tbIn,$tbPseudo,$table,$tb,$tcrpxX,$tcrpxY,$tcdltX,$tcdltY);
  my ($temp1,$temp2,$tlmin,$tlmax);
  my ($v,$verb,$xyunits,$xyunitsString);
  my @dset=();
  my @inimgdim=();
  my @radius_ext;
  my @off_arcm;
  my @pseudoStrAtt=("FILTER","INSTRUME","TELESCOP","AVRG_PNT","RADECSYS","REFYCUNI","REFXCUNI","REFYCTYP");
  my @pseudoR32Att=("PA_PNT","RA_PNT","DEC_PNT");
  push(@pseudoR32Att,"EQUINOX","RA_NOM","DEC_NOM","RA_OBJ","DEC_OBJ");
  push(@pseudoR32Att,"REFYCUNI","REFYDMAX","REFYDMIN","REFYLMAX","REFYLMIN");
  push(@pseudoR32Att,"REFYCDLT","REFYCRVL","REFYCRPX","REFYCTYP","REFXCUNI");
  push(@pseudoR32Att,"REFXDMAX","REFXDMIN","REFXLMAX","REFXLMIN");
  my @pseudoStrAttCol=("TCUNI","TCTYP");
  my @pseudoR32AttCol=("TCDLT","TCRVL","TCRPX","TDMAX","TDMIN","TLMAX","TLMIN");

# ------------------------------------------------------ 
# Read input parameters
# ------------------------------------------------------ 

  $srclist=stringParameter("sourcelistset");
  $evtlist=stringParameter("set");
  $img=stringParameter("imageset");
  $expm=stringParameter("exposureset");
  $flatimg=stringParameter("tmpflatset");
  $temp1=stringParameter("tmpsrcregionset");
  $temp2=stringParameter("tmpbkgregionset");
  $posit_mask=stringParameter("tmpposmaskset");
  $negat_mask=stringParameter("tmpnegmaskset");
  $spec_pos=stringParameter("tmpposspecset");
  $spec_neg=stringParameter("tmpnegspecset");
  $temp3=stringParameter("tmpdiffset");
  $binsize=realParameter("binsize");
  $flspecfile=stringParameter("spectrumset");

#--------------------------------------------------------
# Find out X,Y pixel size in event file
#--------------------------------------------------------
  $ds= DAL::dataset->new(name => $evtlist, accessmode => $DAL::READ);
  $tb= $ds->table(name => "EVENTS");
  
  $co=$tb->column(name=>"X");
  $xyunitsString=$co->units();
  chomp($xyunitsString);
  $xyunitsString=~ s/^\s+//;
  ($xyunits,$pp)=split(/ /,$xyunitsString);

  $at = $co->attribute(name => "TCRPX");
  $tcrpxX=$at->asReal32();
  chomp($tcrpxX);
  $co->release();
  
  $co=$tb->column(name=>"Y");
  $at = $co->attribute(name => "TCRPX");
  $tcrpxY=$at->asReal32();
  chomp($tcrpxY);
  $co->release();

  $co=$tb->column(name=>"X");
  $at = $co->attribute(name => "TCDLT");
  $tcdltX=$at->asReal32();
  chomp($tcdltX);
  $co->release();

  $co=$tb->column(name=>"Y");
  $at = $co->attribute(name => "TCDLT");
  $tcdltY=$at->asReal32();
  chomp($tcdltY);
  $co->release();



  $ds->release();
# ------------------------------------------------------ 
#     Check : input FITS files are EVENT_LIST/SRCLIST/IMAGE/EXPOSURE MAP
# ------------------------------------------------------ 
  
  $ds= DAL::dataset->new(name => $srclist, accessmode => $DAL::READ);
  $tb=$ds->table(name=>"SRCLIST");
  $ds->release();
  
  $ds= DAL::dataset->new(name => $expm, accessmode => $DAL::READ);
  $at = $ds->attribute(name => "CONTENT");
  $key=$at->asString();
  
  unless ($key =~ /EPIC\s*EXPOSURE\s*MAP\s*/){
	SAS::error("NotExpMap","Input image:\"$expm\", is not an EPIC EXPOSURE MAP file\n");
      }
  $ds->release();
  
  $ds= DAL::dataset->new(name => $img, accessmode => $DAL::READ);
  $at = $ds->attribute(name => "CONTENT");
  $ar= $ds->array(name=> 'PRIMARY');
  $key=$at->asString();
  
  unless($key =~ /EPIC\s*IMAGE/){
    SAS::error("NotImg","Input image:\"$img\", is not an EPIC IMAGE file\n");
  }
  
# ------------------------------------------------------ 
#       Obtain some pixel size/position related attributes from imageset
# ------------------------------------------------------ 

  if($ds->hasAttribute(name=>"CDELT1")){
    $at = $ds->attribute(name => "CDELT1");	
  }else{
    $at = $ar->attribute(name => "CDELT1");
  }
  $inPixSize=$at->asReal32();
  
  if($ds->hasAttribute(name=>"CRPIX1")){
    $at = $ds->attribute(name => "CRPIX1");	
  }else{
    $at = $ar->attribute(name => "CRPIX1");
  }
  $crpix1=$at->asReal32();
  
  if($ds->hasAttribute(name=>"CRPIX2")){
    $at = $ds->attribute(name => "CRPIX2");	
  }else{
    $at = $ar->attribute(name => "CRPIX2");
  }
  $crpix2=$at->asReal32();
  

# ------------------------------------------------------ 
#     Calculate INPUT IMAGE dimensions
# ------------------------------------------------------ 
    
  $ar= $ds->array(name=> 'PRIMARY');
  @inimgdim[0,1]=$ar->dimensions();
  chomp($inimgdim[0]);
  $inimgdim[0] =~ s/\s+//g;
  $inimgdim[1] =~ s/\s+//g;
  chomp($inimgdim[1]);

  $ds->release();

# ------------------------------------------------------ 
# RUN Flspec ...
# ------------------------------------------------------ 

#
# ------------------------------------------------------ 
# 1) Write regions for ALL sources through task: REGION
#    Input:  SRC_LIST_file
#            
#    Output: SRC_REGION_file --> temp1
# ------------------------------------------------------ 
#
  $srclist=$srclist.":SRCLIST";
  
  $msg="Writting regions for all sources in source list...\n";
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);
  
  if(system("region eventset=$evtlist sourcelisttable=$srclist operationmode='global' regionfile=$temp1 bkgregionfile=$temp2 radiusmode='enfrac' energyfraction=0.95 extractionmode='local'")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrRegion","Cannot run REGION on \"$srclist\"\n");
  }

  (-e $temp2) && unlink($temp2);
    
# ------------------------------------------------------ 
#  2) Create a mask to "mask" source regions in input image
#---------------------------------------------------------
#
#     2.1) Create pseudo evt list with X/Y/DETX/DETY/THETA columns
#          (run createpseudoevt F90 module) --> temp2
#
  if(system("createpseudoevt set=$temp2 imageset=$img tcrpxx=$tcrpxX tcrpxy=$tcrpxY tcdltx=$tcdltX tcdlty=$tcdltY")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrPseudo","Cannot create Pseudo EVT");
  }
  $table=$temp2.":EVENTS";


#
#     2.2) Write WCS keywords to PSEUDO event list
#

  $evIn= DAL::dataset->new(name => $evtlist, accessmode => $DAL::READ);
  $tbIn=$evIn->table(name=>"EVENTS");
  $coXIn=$tbIn->column(name=>"X");
  $coYIn=$tbIn->column(name=>"Y");

  $evPseudo= DAL::dataset->new(name => $temp2, accessmode => $DAL::MODIFY);
  $tbPseudo=$evPseudo->table(name=>"EVENTS");
  $coXPseudo=$tbPseudo->column(name=>"X");
  $coYPseudo=$tbPseudo->column(name=>"Y");

  foreach $att (@pseudoStrAtt){
    $at = $tbIn->attribute(name => $att);
    $key=$at->asString();
    $evPseudo->addAttribute(name=>$att,datatype=>$DAL::STRING,value=>$key);
    if($att =~ /INSTRUME/){
      if($key =~ /EMOS/){
	$specBin=15;
	$specChanMax=11999;
      }elsif($key =~ /EPN/){
	$specBin=5;
	$specChanMax=20479;
      }
    }
  }
    
  foreach $att (@pseudoR32Att){
    $at = $tbIn->attribute(name => $att);
    $key=$at->asReal32();
    $evPseudo->addAttribute(name=> $att,datatype=>$DAL::REAL32,value=>$key);
  }


  foreach $att (@pseudoStrAttCol){
    $at = $coXIn->attribute(name => $att);
    $key=$at->asString();
    $coXPseudo->addAttribute(name=> $att,datatype=>$DAL::STRING,value=>$key);
    $at = $coYIn->attribute(name => $att);
    $key=$at->asString();
    $coYPseudo->addAttribute(name=> $att,datatype=>$DAL::STRING,value=>$key);
  }

  foreach $att (@pseudoR32AttCol){
    if($att =~ /TCRPX/){
      $key=$tcrpxX;
    }else{
      $at = $coXIn->attribute(name => $att);
      $key=$at->asReal32();
    }
    $coXPseudo->addAttribute(name=> $att,datatype=>$DAL::REAL32,value=>$key);
    if($att =~ /TCRPX/){
      $key=$tcrpxY;
    }else{
      $at = $coYIn->attribute(name => $att);
      $key=$at->asReal32();
    }
    $coYPseudo->addAttribute(name=> $att,datatype=>$DAL::REAL32,value=>$key);
  }

  $evIn->release();
  $evPseudo->release();

#
#     2.3) Filter pseudo evt list with region output and save image =mask
#          ---> temp3


  if(system("evselect table=$table  expression='region($temp1,X,Y)' withfilteredset=false keepfilteroutput=false destruct=false writedss=false updateexposure=false withimageset=true imageset=$temp3 xcolumn='X' ycolumn='Y' imagebinning=binSize ximagebinsize=80 yimagebinsize=80")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrEvselect","Cannot run EVSELECT to filter input evt list with REGION output\n");
  }


#
#     2.3) Modify mask so that it has 0 in source positions and 1 otherwise 
#          --> temp1 -->temp3
#
  if(system("fcarith infile=$temp3 outfil=$temp1 ops=MUL const='-1' datatype='-' copyprime=yes clobber=yes")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFcarith","Cannot calculate regionmask * (-1)\n");
  }
  unlink($temp3);
  if(system("fcarith infile=$temp1 outfil=$temp3 ops=ADD const='1' datatype='-' copyprime=yes clobber=yes")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFcarith","Cannot calculate regionmask*(-1)+1\n");
  }
  unlink($temp1);

#
# ------------------------------------------------------ 
#  3) Flatfield input IMAGE with its EXPOSMAP and mask out the sources
#       Input:  IMG_file
#               EXPOSMAP_file
# 
#       Output: FLAT_IMG_file
# ------------------------------------------------------ 
#
  #
  #   3.0) create mask to "clean" borders of expmap
  #
  if(system("emask expimageset=$expm detmaskset=$flatimg threshold1=0.1")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrEmask","Cannot create mask for \"$expm\"\n");
  }
  if(system("farith infil1=$expm infil2=$flatimg outfil=$temp1 ops=MUL datatype='-' blank=-1 null=no copyprime=yes clobber=yes")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFarith","Cannot multiply: exposure_mask and exposure_map\n");
  }
  (-e $flatimg) &&  unlink($flatimg);

  #
  #   3.1) imageset/exposuremap_masked= flatimg
  #
  
  $msg="Flatfielding input image with its corresponding exposure map...\n";
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);
  
  if(system("farith infil1=$img infil2=$temp1 outfil=$flatimg ops=DIV datatype='-' blank=-1 null=no copyprime=yes clobber=yes")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFarith","Cannot divide:\"$img\" and \"$expm\"\n");
  }
  #
  #   3.2) flatimg/region mask (to hide sources) --> temp1
  #
  
  if(system("farith infil1=$flatimg infil2=$temp3 outfil=$temp1 ops=DIV datatype='-' blank=-1 null=no copyprime=yes clobber=yes")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFarith","Cannot calculate (flatfield/regionmask\n");
  }
  
  unlink($flatimg);
  unlink($temp3);

#
# ------------------------------------------------------ 
#  4) Create "median" mask file to select pixels above/below 
#     the "median": FLMASK
#      (F90 module program)
#      Input:  FLAT_IMG_file
#
#      Output: MASK_POS_file
#              MASK_NEG_file
# ------------------------------------------------------
#

  $msg="Creating mask files...\n";
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);
  
  if(system("flmask set=$evtlist imageset=$temp1 binsize=$binsize positivemaskset=$posit_mask negativemaskset=$negat_mask")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrFlmask","Mask files cannot be created\n");
  }
  unlink($temp1);

# ------------------------------------------------------ 
#  5) Establish same lower/upper channel values for all 
#     the spectra 
#     
# ------------------------------------------------------ 

  if(system("fstatistic infile=$evtlist colname=PI rows='-' maxval=INDEF minval=INDEF clobber=yes outfile=$temp1")){
    clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
    clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
    SAS::error("ErrStat","Cannot calculate statistic for \"$evtlist\"\n"); 
  }
  unlink($temp1);
  
  $ds= DAL::dataset->new(name => $evtlist, accessmode => $DAL::READ);
  $tb= $ds->table(name => "EVENTS");
  $tlmin=1;
  $tlmax=1;
  
  if($tb->hasColumn(name=>"PI")){
    $co=$tb->column(name=>"PI");
    if($co->hasAttribute(name=>"TLMAX")){
      $at=$co->attribute(name=>"TLMAX");
      $tlmax=$at->asReal32();    
    }
    if($co->hasAttribute(name=>"TLMIN")){
      $at=$co->attribute(name=>"TLMIN");
      $tlmin=$at->asReal32();    
    }
  }
# if($tlmax > $specChanMax){$tlmax=$specChanMax;}
  $hascol=0;
  if($tb->hasColumn(name=>"CCDNR")){
    $hascol=1;
  }
  $tb->release();    
  $ds->release();
  if($tlmax == 1){$tlmax=`pget fstatistic max`;}
    
    
#
# ------------------------------------------------------ 
#  6) Define annular regions for background spectra. 
# ------------------------------------------------------ 
#
  for($jj=0;$jj<$nann;$jj++){
    $jj1=$jj+1;
    $msg="Spatial region number: $jj1\n";
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);
    
    $radius_ext[$jj]=sqrt(($jj+1)/$nann)*$rmax * 60.0; # arcs
    
    if($jj==0){
      $filter="'"."THETA <= $radius_ext[$jj]"."'";
      $off_arcs=$radius_ext[$jj]/2.0;
      $radius_int=0.0;
    }else{
      $radius_int=$radius_ext[$jj-1];
      $filter="'"."THETA > $radius_int && THETA <= $radius_ext[$jj]"."'";
      $off_arcs=($radius_ext[$jj]+$radius_ext[$jj-1])/2.0;
    }
    $off_arcm[$jj]=$off_arcs/60.0;
    
#-------------------------------------------------------------
#    7) Filter pseudo evt list to obtain annulus_mask:$temp1
#-------------------------------------------------------------

    if(system("evselect table=$table  expression=$filter withfilteredset=false keepfilteroutput=false destruct=false writedss=false updateexposure=false withimageset=true imageset=$temp1 xcolumn='X' ycolumn='Y' imagebinning=binSize ximagebinsize=80 yimagebinsize=80")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrEvselect","Cannot filter Pseudo EVT to obtain annulusmask\n");
    }

#
# ------------------------------------------------------ 
#  8) Filter EVT_LIST with POS/NEG MASKs * annular mask
#     and extract "positive/negative" spectra : EVSELECT
#     Input:  EVT_LIST_file
#             MASK_POS(NEG)_mask*ANNULAR_mask
#        
#     Output: SPEC_POS(NEG)_file_in_annulus
# ------------------------------------------------------ 
#
	
    #
    #    8.1) Do annulus_mask*posit_mask=final posit_mask -->temp3
    #         annulus_mask has CREATOR & DATE attributes
    $filterexp="'mask("."$temp3".",0,0,X,Y)'";
    $evt=$evtlist.":EVENTS";
    
    $posit1=$posit_mask."+1";
    if(system("farith infil1=$temp1 infil2=$posit1 outfil=$temp3 ops=MUL datatype='-' blank=-1 null=no copyprime=yes clobber=yes")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFarith","Cannot calculate posmask*annulusmask\n");
    }


    #
    # Add a couple of keywords so that evselect does not complain
    #

    $msg="Filtering events and extracting \"positive\" spectrum...\n";
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);

    #
    #    8.2) Extract positive spectrum -->spec_pos
    #
    if(system("evselect table=$evt  expression=$filterexp withfilteredset=false keepfilteroutput=false destruct=false writedss=true updateexposure=true withspectrumset=true spectrumset=$spec_pos energycolumn=PI spectralbinsize=$specBin withspecranges=true specchannelmin=$tlmin specchannelmax=$tlmax")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrEvselect","Cannot create Positive spectrum\n");
    }
    #
    #     8.3) Obtain Solid Angle for positive spectra: count "1" pixels 
    #          in filter mask
    #

    if(system("fimgstat infile=$temp3 threshlo=1 threshup=1 outfile=$flatimg clobber=yes")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFimgstat","Cannot run fimgstat to obtain num.of pixels>0\n");
    }

    unlink($flatimg);
    $SolidAngle_pos=`pget fimgstat num`;
    $SolidAngle_pos=$SolidAngle_pos*($inPixSize/$tcdltX)**2;
    unlink($temp3);
    
    #
    #    8.4) Do annulus_mask*neg_mask=final neg_mask -->temp3
    #      
    $negat1=$negat_mask."+1";
    if(system("farith infil1=$temp1 infil2=$negat1 outfil=$temp3 ops=MUL datatype='-' blank=-1 null=no copyprime=yes clobber=yes")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFarith","Cannot calculate negmask*annulusmask\n");
    }

    $msg="Filtering events and extracting \"negative\" spectrum...\n";
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);
    
    #
    #    8.5) Extract negative spectrum -->spec_neg
    #
    if(system("evselect table=$evt  expression=$filterexp withfilteredset=false keepfilteroutput=false destruct=false writedss=true updateexposure=true withspectrumset=true spectrumset=$spec_neg energycolumn=PI spectralbinsize=$specBin withspecranges=true specchannelmin=$tlmin specchannelmax=$tlmax")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrEvselect","Cannot create Negative spectrum\n");
      
    }
    #
    #    8.6) Obtain Solid Angle for negative spectra: 
    #         count "1" pixels in filter mask
    #
    if(system("fimgstat infile=$temp3 threshlo=1 threshup=1 outfile=$flatimg clobber=yes")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFimgstat","Cannot run fimgstat to obtain num.of pixels>0");
    }
    unlink($flatimg);
    $SolidAngle_neg=`pget fimgstat num`;
    $SolidAngle_neg=$SolidAngle_neg*($inPixSize/$tcdltX)**2;
    unlink($temp1);
    unlink($temp3);
#
# ------------------------------------------------------ 
#   9) Subtract SPEC_NEG spectrum from SPEC_POS spectrum: MATHPHA
#       Input:  SPEC_NEG_file_in_annulus
#               SPEC_POS_file_in_annulus
#        
#       Output: DIFF_SPEC_file_in annulus
# ------------------------------------------------------ 

    $nspec=$jj+1;
    
    #
    #  9.1)Normalise both spectra to Positive spectrum solid angle
    #
    
    
    $msg="Normalising both spectra to Positive spectrum solid angle...\n";
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);
    
    if($SolidAngle_pos == 0. && $SolidAngle_neg > 0.){
      $expr="'0.0*".$spec_pos."-(".$spec_neg.")'";
      $SolidAngle=$SolidAngle_neg;
      SAS::warning("NullSpecPos","Positive spectrum has null extraction area for offaxis=$off_arcm[$jj] (spectrum number $nspec) \n");	  
      
    }elsif($SolidAngle_pos > 0. && $SolidAngle_neg == 0.){
      $expr="'".$spec_pos."-0.0*".$spec_neg."'";
      $SolidAngle=$SolidAngle_pos;
      SAS::warning("NullSpecNeg","Negative spectrum has null extraction area for offaxis=$off_arcm[$jj] (spectrum number $nspec) \n");	  
      
    }elsif($SolidAngle_pos <= 0. || $SolidAngle_neg <= 0.){
      $SolidAngle=0.0;
      $expr="'0.0*".$spec_pos."-".$spec_neg."*0.0'";
      SAS::warning("NullSpec","Positive and Negative spectrum have null extraction area for offaxis=$off_arcm[$jj] (spectrum number $nspec) \n");
    }else{	
      $factor=$SolidAngle_pos/$SolidAngle_neg;
      $factorFormat=sprintf "%.3f",$factor;
      $expr="'".$spec_pos."-(".$spec_neg."*".$factorFormat.")'";
      $SolidAngle=$SolidAngle_pos;
    }
    
    
    $SolidAngleFormat=sprintf "%.5e",$SolidAngle;
    #
    #  9.2) EXPOSURE update just in case evselect cannot do it 
    #       (MATHPHA cannot operate without this keyword in datasets)
    #
    
    $dset[0]=$spec_pos;
    $dset[1]=$spec_neg;
    
    for($ids=0;$ids<2;$ids++){
      $exposds=0.000;
      $expostb=0.000;
      $ds= DAL::dataset->new(name => $dset[$ids], accessmode => $DAL::MODIFY);
      $tb= $ds->table(name=>"SPECTRUM");
      if($ds->hasAttribute(name => "EXPOSURE")){
	$at = $ds->attribute(name => "EXPOSURE");
	$exposds=$at->asReal32();
	if($tb->hasAttribute(name => "EXPOSURE")){
	  $at = $tb->attribute(name => "EXPOSURE");
	  $expostb=$at->asReal32();
	}else{
	  $tb->addAttribute(name => "EXPOSURE", datatype =>$DAL::REAL32, value=>$expostb);	
	  
	}
      }elsif($tb->hasAttribute(name => "EXPOSURE")){
	$at = $tb->attribute(name => "EXPOSURE");
	$expostb=$at->asReal32();
	
      }else{
	SAS::warning("NullExpos1","Exposure value for spectrum $dset[$ids] cannot be properly updated by evselect in fluctuations spectrum number $nspec: Exposure updated to 1.000 \n");	  
	$expostb=1.000;
	$tb->addAttribute(name => "EXPOSURE", datatype =>$DAL::REAL32, value=>$expostb);	
	
      }
      if($expostb==0.0){
	if($exposds != 0.0){
	  $expostb=$exposds;
	  SAS::warning("NullExposDs","Exposure value for spectrum $dset[$ids] cannot be properly updated by evselect in fluctuations spectrum number $nspec: Exposure updated to EXPOSURE value in event list dataset \n");	  
	}else{
	  SAS::warning("NullExpos1","Exposure value for spectrum $dset[$ids] cannot be properly updated by evselect in fluctuations spectrum number $nspec: Exposure updated to 1.000 \n");	  
	  $expostb=1.0000;
	}
	$tb->deleteAttribute(name => "EXPOSURE");
	$tb->addAttribute(name => "EXPOSURE", datatype =>$DAL::REAL32, value=>$expostb);
      }
      $ds->release();
    }
    
    #
    #   9.3) Use MATHPHA to calculate the difference spectrum
    # 
    
    $verb=$ENV{"SAS_VERBOSITY"}-1;
    
    $msg="Calculating difference spectrum...\n";
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);
    
    $msg="Executing (routine):mathpha expr=$expr units='C' outfil=$temp3 exposure=$spec_pos Areascal='%' properr='yes' errmeth='POISS-1' ncomments=0 auxfiles='NONE' backfile='NONE' backscal=$SolidAngleFormat corrfile='NONE' corrscal='NONE' arfile='NONE' rmfile='NONE' phaversn='1.1.0' divzero=-99 chatter=$verb  clobber=yes\n";

    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);
    
    if(system("mathpha expr=$expr units='C' outfil=$temp3 exposure=$spec_pos Areascal='%' properr='yes' errmeth='POISS-1' ncomments=0 auxfiles='NONE' backfile='NONE' backscal=$SolidAngleFormat corrfile='NONE' corrscal='NONE' arfile='NONE' rmfile='NONE' phaversn='1.1.0' divzero=-99 chatter=$verb  clobber=yes")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrMathpha","MATHPHA: Subtracted spectrum cannot be created\n");
    }
    
    $temp31=$temp3."+1";
    if(system("fparkey value=0 fitsfile=$temp31 keyword=-CREATOR")){
      clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
      clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
      SAS::error("ErrFparkey","Cannot run fparkey to remove CREATOR attribute");
    }
    (-e $spec_neg) && unlink($spec_neg);
    (-e $spec_pos) && unlink($spec_pos);
    
    #
    #  9.4) Add difference spectrum to final spectrum (new row):SPECADD
    #
    
    $msg="Adding difference spectrum as a new row in output spectral file...\n";
    SAS::message($SAS::AppMsg,$SAS::VerboseMsg,$msg);
    
    if($nspec == 1){
      if(system("specadd sets=$temp3 useoldsumset=no sumset=$flspecfile")){
	clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
	clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
	SAS::error("ErrFladdspec","Cannot store single spectrum\n");
      }
    }else{
      # rename multiple file to temp file to store one more single spectrum
      rename($flspecfile,$spec_pos);
      if(system("specadd sets=$temp3 useoldsumset=yes oldsumset=$spec_pos sumset=$flspecfile")){
	clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
	clearTemporary($spec_pos,$spec_neg,$temp3,$flspecfile);
	SAS::error("ErrFladdspec","Cannot store single spectrum\n");
      }
    }
    (-e $spec_pos)&& unlink($spec_pos);
    (-e $temp3)&& unlink($temp3);
  } #end offaxis values
  
#------------------------------------------------------
#   10) Save offaxis values in ROWID column of multiple spectra
#------------------------------------------------------

  $ds= DAL::dataset->new(name => $flspecfile, accessmode => $DAL::MODIFY);
  $tb= $ds->table(name => "SPECTRUM");
  
  $co=$tb->column(name=>"ROWID");
  $v=$co->data();
  for($jj=0;$jj<$nann;$jj++){
    $rowidStr=sprintf "%.5f",$off_arcm[$jj];
    $rowidStr .= "   arcm";
    $v->set($rowidStr,$jj*80);
  }

 $co->release();

#------------------------------------------------------
#   11) Add some attributes
#------------------------------------------------------
#  
  $label="Epic Fluctuations Spectra at 10 different off-axis positions specified in ROWID column";
  $ds->addAttribute(name=>"CONTENT",datatype=>$DAL::STRING,value=>"EPIC FLUCTUATIONS SPECTRUM",units=>'',label=>$label);
  
  $label="Multiple PHA file contained";
  $tb->addAttribute(name=>"HDUCLAS3",datatype=>$DAL::STRING,value=>"COUNT",units=>'',label=>$label);
  
  $ds->release();
  
# ------------------------------------------------------ 
#   12) Remove intermediate files
# ------------------------------------------------------ 
#
  clearTemporary($temp1,$temp2,$temp3,$posit_mask,$negat_mask,$flatimg);
  clearTemporary($spec_pos,$spec_neg,$temp3);
  
}

sub clearTemporary{
  foreach $_ (@_){
    (-e $_) && unlink($_);
  }
}
